import torch
from torch.nn import Module, Linear, LayerNorm, Dropout
from transformers import BertPreTrainedModel, AutoModel
from transformers.activations import ACT2FN

from fastcoref.utilities.util import extract_clusters, extract_mentions_to_clusters, mask_tensor

# took from: https://github.com/yuvalkirstain/s2e-coref


class FullyConnectedLayer(Module):
    def __init__(self, config, input_dim, output_dim, dropout_prob):
        super(FullyConnectedLayer, self).__init__()

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.dropout_prob = dropout_prob

        self.dense = Linear(self.input_dim, self.output_dim)
        self.layer_norm = LayerNorm(self.output_dim, eps=config.layer_norm_eps)
        self.activation_func = ACT2FN[config.hidden_act]
        self.dropout = Dropout(self.dropout_prob)

    def forward(self, inputs):
        temp = inputs
        temp = self.dense(temp)
        temp = self.activation_func(temp)
        temp = self.layer_norm(temp)
        temp = self.dropout(temp)
        return temp


class FCorefModel(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.max_span_length = config.coref_head['max_span_length']
        self.top_lambda = config.coref_head['top_lambda']
        self.ffnn_size = config.coref_head['ffnn_size']
        self.dropout_prob = config.coref_head['dropout_prob']

        base_model = AutoModel.from_config(config)
        FCorefModel.base_model_prefix = base_model.base_model_prefix
        FCorefModel.config_class = base_model.config_class
        setattr(self, self.base_model_prefix, base_model)

        self.start_mention_mlp = FullyConnectedLayer(config, config.hidden_size, self.ffnn_size, self.dropout_prob)
        self.end_mention_mlp = FullyConnectedLayer(config, config.hidden_size, self.ffnn_size, self.dropout_prob)

        self.start_coref_mlp = FullyConnectedLayer(config, config.hidden_size, self.ffnn_size, self.dropout_prob)
        self.end_coref_mlp = FullyConnectedLayer(config, config.hidden_size, self.ffnn_size, self.dropout_prob)

        self.mention_start_classifier = Linear(self.ffnn_size, 1)
        self.mention_end_classifier = Linear(self.ffnn_size, 1)
        self.mention_s2e_classifier = Linear(self.ffnn_size, self.ffnn_size)

        self.antecedent_s2s_classifier = Linear(self.ffnn_size, self.ffnn_size)
        self.antecedent_e2e_classifier = Linear(self.ffnn_size, self.ffnn_size)
        self.antecedent_s2e_classifier = Linear(self.ffnn_size, self.ffnn_size)
        self.antecedent_e2s_classifier = Linear(self.ffnn_size, self.ffnn_size)

        self.init_weights()

    def num_parameters(self) -> tuple:
        def head_filter(x):
            return x[1].requires_grad and any(hp in x[0] for hp in ['coref', 'mention', 'antecedent'])

        head_params = filter(head_filter, self.named_parameters())
        head_params = sum(p.numel() for n, p in head_params)
        return super().num_parameters() - head_params, head_params

    def _get_span_mask(self, batch_size, k, max_k):
        """
        :param batch_size: int
        :param k: tensor of size [batch_size], with the required k for each example
        :param max_k: int
        :return: [batch_size, max_k] of zero-ones, where 1 stands for a valid span and 0 for a padded span
        """
        size = (batch_size, max_k)
        idx = torch.arange(max_k, device=self.device).unsqueeze(0).expand(size)
        len_expanded = k.unsqueeze(1).expand(size)
        return (idx < len_expanded).int()

    def _prune_topk_mentions(self, mention_logits, attention_mask, topk_1d_indices):
        """
        :param mention_logits: Shape [batch_size, seq_length, seq_length]
        :param attention_mask: [batch_size, seq_length]
        :param top_lambda:
        :return:
        """
        batch_size, seq_length, _ = mention_logits.size()
        actual_seq_lengths = torch.sum(attention_mask, dim=-1)  # [batch_size]

        k = (actual_seq_lengths * self.top_lambda).int()  # [batch_size]
        max_k = int(torch.max(k))  # This is the k for the largest input in the batch, we will need to pad

        if topk_1d_indices is None:
            _, topk_1d_indices = torch.topk(mention_logits.view(batch_size, -1), dim=-1, k=max_k)  # [batch_size, max_k]

        span_mask = self._get_span_mask(batch_size, k, max_k)  # [batch_size, max_k]
        # drop the invalid indices and set them to the last index
        topk_1d_indices = (topk_1d_indices * span_mask) + (1 - span_mask) * ((seq_length ** 2) - 1)  # We take different k for each example
        # sorting for coref mention order
        sorted_topk_1d_indices, _ = torch.sort(topk_1d_indices, dim=-1)  # [batch_size, max_k]

        # gives the row index in 2D matrix
        topk_mention_start_ids = torch.div(sorted_topk_1d_indices, seq_length, rounding_mode='floor') # [batch_size, max_k]
        topk_mention_end_ids = sorted_topk_1d_indices % seq_length  # [batch_size, max_k]

        topk_mention_logits = mention_logits[torch.arange(batch_size).unsqueeze(-1).expand(batch_size, max_k),
                                             topk_mention_start_ids, topk_mention_end_ids]  # [batch_size, max_k]

        # this is antecedents scores - rows mentions, cols coref mentions
        topk_mention_logits = topk_mention_logits.unsqueeze(-1) + topk_mention_logits.unsqueeze(-2)  # [batch_size, max_k, max_k]

        return topk_mention_start_ids, topk_mention_end_ids, span_mask, topk_mention_logits

    def _mask_antecedent_logits(self, antecedent_logits, span_mask):
        # We now build the matrix for each pair of spans (i,j) - whether j is a candidate for being antecedent of i?
        antecedents_mask = torch.ones_like(antecedent_logits, dtype=self.dtype).tril(diagonal=-1)  # [batch_size, k, k]
        antecedents_mask = antecedents_mask * span_mask.unsqueeze(-1)  # [batch_size, k, k]
        antecedent_logits = mask_tensor(antecedent_logits, antecedents_mask)
        return antecedent_logits

    def _get_cluster_labels_after_pruning(self, span_starts, span_ends, all_clusters):
        """
        :param span_starts: [batch_size, max_k]
        :param span_ends: [batch_size, max_k]
        :param all_clusters: [batch_size, max_cluster_size, max_clusters_num, 2]
        :return: [batch_size, max_k, max_k + 1] - [b, i, j] == 1 if i is antecedent of j
        """
        batch_size, max_k = span_starts.size()
        new_cluster_labels = torch.zeros((batch_size, max_k, max_k + 1), device='cpu')
        all_clusters_cpu = all_clusters.cpu().numpy()
        for b, (starts, ends, gold_clusters) in enumerate(zip(span_starts.cpu().tolist(), span_ends.cpu().tolist(), all_clusters_cpu)):
            gold_clusters = extract_clusters(gold_clusters)
            mention_to_gold_clusters = extract_mentions_to_clusters(gold_clusters)
            gold_mentions = set(mention_to_gold_clusters.keys())
            for i, (start, end) in enumerate(zip(starts, ends)):
                if (start, end) not in gold_mentions:
                    continue
                for j, (a_start, a_end) in enumerate(list(zip(starts, ends))[:i]):
                    if (a_start, a_end) in mention_to_gold_clusters[(start, end)]:
                        new_cluster_labels[b, i, j] = 1
        new_cluster_labels = new_cluster_labels.to(self.device)
        no_antecedents = 1 - torch.sum(new_cluster_labels, dim=-1).bool().float()
        new_cluster_labels[:, :, -1] = no_antecedents
        return new_cluster_labels

    def _get_marginal_log_likelihood_loss(self, coref_logits, cluster_labels_after_pruning, span_mask):
        """
        :param coref_logits: [batch_size, max_k, max_k]
        :param cluster_labels_after_pruning: [batch_size, max_k, max_k]
        :param span_mask: [batch_size, max_k]
        :return:
        """
        gold_coref_logits = mask_tensor(coref_logits, cluster_labels_after_pruning)

        gold_log_sum_exp = torch.logsumexp(gold_coref_logits, dim=-1)  # [batch_size, max_k]
        all_log_sum_exp = torch.logsumexp(coref_logits, dim=-1)  # [batch_size, max_k]

        gold_log_probs = gold_log_sum_exp - all_log_sum_exp
        losses = - gold_log_probs

        losses = losses * span_mask
        per_example_loss = torch.sum(losses, dim=-1)  # [batch_size]

        per_example_loss = per_example_loss / losses.size(-1)
        loss = per_example_loss.mean()
        return loss

    def _get_mention_mask(self, mention_logits_or_weights):
        """
        Returns a tensor of size [batch_size, seq_length, seq_length] where valid spans
        (start <= end < start + max_span_length) are 1 and the rest are 0
        :param mention_logits_or_weights: Either the span mention logits or weights, size [batch_size, seq_length, seq_length]
        """
        mention_mask = torch.ones_like(mention_logits_or_weights, dtype=self.dtype)
        mention_mask = mention_mask.triu(diagonal=0)
        mention_mask = mention_mask.tril(diagonal=self.max_span_length - 1)
        return mention_mask

    def _calc_mention_logits(self, start_mention_reps, end_mention_reps):
        start_mention_logits = self.mention_start_classifier(start_mention_reps).squeeze(-1)  # [batch_size, seq_length]
        end_mention_logits = self.mention_end_classifier(end_mention_reps).squeeze(-1)  # [batch_size, seq_length]

        temp = self.mention_s2e_classifier(start_mention_reps)  # [batch_size, seq_length]
        joint_mention_logits = torch.matmul(temp,
                                            end_mention_reps.permute([0, 2, 1]))  # [batch_size, seq_length, seq_length]

        mention_logits = joint_mention_logits + start_mention_logits.unsqueeze(-1) + end_mention_logits.unsqueeze(-2)
        mention_mask = self._get_mention_mask(mention_logits)  # [batch_size, seq_length, seq_length]
        mention_logits = mask_tensor(mention_logits, mention_mask)  # [batch_size, seq_length, seq_length]
        return mention_logits

    def _calc_coref_logits(self, top_k_start_coref_reps, top_k_end_coref_reps):
        # s2s
        temp = self.antecedent_s2s_classifier(top_k_start_coref_reps)  # [batch_size, max_k, dim]
        top_k_s2s_coref_logits = torch.matmul(temp,
                                              top_k_start_coref_reps.permute([0, 2, 1]))  # [batch_size, max_k, max_k]

        # e2e
        temp = self.antecedent_e2e_classifier(top_k_end_coref_reps)  # [batch_size, max_k, dim]
        top_k_e2e_coref_logits = torch.matmul(temp,
                                              top_k_end_coref_reps.permute([0, 2, 1]))  # [batch_size, max_k, max_k]

        # s2e
        temp = self.antecedent_s2e_classifier(top_k_start_coref_reps)  # [batch_size, max_k, dim]
        top_k_s2e_coref_logits = torch.matmul(temp,
                                              top_k_end_coref_reps.permute([0, 2, 1]))  # [batch_size, max_k, max_k]

        # e2s
        temp = self.antecedent_e2s_classifier(top_k_end_coref_reps)  # [batch_size, max_k, dim]
        top_k_e2s_coref_logits = torch.matmul(temp,
                                              top_k_start_coref_reps.permute([0, 2, 1]))  # [batch_size, max_k, max_k]

        # sum all terms
        coref_logits = top_k_s2e_coref_logits + top_k_e2s_coref_logits + top_k_s2s_coref_logits + top_k_e2e_coref_logits  # [batch_size, max_k, max_k]
        return coref_logits

    def forward_transformer(self, batch):
        input_ids = batch['input_ids']
        attention_mask = batch['attention_mask']

        docs, segments, segment_len = input_ids.size()
        input_ids, attention_mask = input_ids.view(-1, segment_len), attention_mask.view(-1, segment_len)

        outputs = self.base_model(input_ids, attention_mask=attention_mask)
        sequence_output = outputs.last_hidden_state

        attention_mask = attention_mask.view((docs, segments * segment_len))        # [docs, seq_len]
        sequence_output = sequence_output.view((docs, segments * segment_len, -1))  # [docs, seq_len, dim]

        leftovers_ids, leftovers_mask = batch['leftovers']['input_ids'], batch['leftovers']['attention_mask']
        if len(leftovers_ids) > 0:
            res_outputs = self.base_model(leftovers_ids, attention_mask=leftovers_mask)
            res_sequence_output = res_outputs.last_hidden_state

            attention_mask = torch.cat([attention_mask, leftovers_mask], dim=1)
            sequence_output = torch.cat([sequence_output, res_sequence_output], dim=1)

        return sequence_output, attention_mask

    def forward(self, batch, gold_clusters=None, topk_1d_indices=None, return_all_outputs=False):
        sequence_output, attention_mask = self.forward_transformer(batch)

        # Compute representations
        start_mention_reps = self.start_mention_mlp(sequence_output)
        end_mention_reps = self.end_mention_mlp(sequence_output)

        start_coref_reps = self.start_coref_mlp(sequence_output)
        end_coref_reps = self.end_coref_mlp(sequence_output)

        # mention scores
        mention_logits = self._calc_mention_logits(start_mention_reps, end_mention_reps)

        # prune mentions
        mention_start_ids, mention_end_ids, span_mask, topk_mention_logits = self._prune_topk_mentions(mention_logits, attention_mask, topk_1d_indices)

        batch_size, _, dim = start_coref_reps.size()
        max_k = mention_start_ids.size(-1)
        size = (batch_size, max_k, dim)

        # Antecedent scores
        # gather reps
        topk_start_coref_reps = torch.gather(start_coref_reps, dim=1, index=mention_start_ids.unsqueeze(-1).expand(size))
        topk_end_coref_reps = torch.gather(end_coref_reps, dim=1, index=mention_end_ids.unsqueeze(-1).expand(size))
        coref_logits = self._calc_coref_logits(topk_start_coref_reps, topk_end_coref_reps)

        final_logits = topk_mention_logits + coref_logits
        final_logits = self._mask_antecedent_logits(final_logits, span_mask)
        # adding zero logits for null span
        final_logits = torch.cat((final_logits, torch.zeros((batch_size, max_k, 1), device=self.device)), dim=-1)  # [batch_size, max_k, max_k + 1]

        if return_all_outputs:
            outputs = (mention_start_ids, mention_end_ids, mention_logits, final_logits)
        else:
            outputs = tuple()

        if topk_1d_indices is not None:
            outputs = (span_mask,) + outputs

        if gold_clusters is not None:
            labels_after_pruning = self._get_cluster_labels_after_pruning(mention_start_ids, mention_end_ids, gold_clusters)
            loss = self._get_marginal_log_likelihood_loss(final_logits, labels_after_pruning, span_mask)
            outputs = (loss, ) + outputs

        return outputs





import math
import numpy as np
import torch
from torch import nn
from torch.nn import Module, Linear, LayerNorm, Dropout, init
from transformers import BertPreTrainedModel, AutoModel
from transformers.activations import ACT2FN

from fastcoref.utilities.consts import CATEGORIES, STOPWORDS
from fastcoref.utilities.util import extract_clusters, extract_mentions_to_clusters, mask_tensor, get_pronoun_id, get_category_id


class FullyConnectedLayer(Module):
    def __init__(self, config, input_dim, output_dim, dropout_prob):
        super(FullyConnectedLayer, self).__init__()

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.dropout_prob = dropout_prob

        self.dense = Linear(self.input_dim, self.output_dim)
        self.layer_norm = LayerNorm(self.output_dim, eps=config.layer_norm_eps)
        self.activation_func = ACT2FN[config.hidden_act]
        self.dropout = Dropout(self.dropout_prob)

    def forward(self, inputs):
        temp = inputs
        temp = self.dense(temp)
        temp = self.activation_func(temp)
        temp = self.layer_norm(temp)
        temp = self.dropout(temp)
        return temp


class LingMessModel(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.max_span_length = config.coref_head['max_span_length']
        self.top_lambda = config.coref_head['top_lambda']
        self.ffnn_size = config.coref_head['ffnn_size']
        self.dropout_prob = config.coref_head['dropout_prob']
        self.hidden_size = config.hidden_size

        self.num_cats = len(CATEGORIES) + 1                 # +1 for ALL
        self.all_cats_size = self.ffnn_size * self.num_cats

        # this is how huggingface loading the class model and setting the name of the variable.
        base_model = AutoModel.from_config(config)
        LingMessModel.base_model_prefix = base_model.base_model_prefix
        LingMessModel.config_class = base_model.config_class
        setattr(self, self.base_model_prefix, base_model)

        self.start_mention_mlp = FullyConnectedLayer(config, self.hidden_size, self.ffnn_size, self.dropout_prob)
        self.end_mention_mlp = FullyConnectedLayer(config, self.hidden_size, self.ffnn_size, self.dropout_prob)

        self.mention_start_classifier = Linear(self.ffnn_size, 1)
        self.mention_end_classifier = Linear(self.ffnn_size, 1)
        self.mention_s2e_classifier = Linear(self.ffnn_size, self.ffnn_size)

        self.coref_start_all_mlps = FullyConnectedLayer(config, config.hidden_size, self.all_cats_size, self.dropout_prob)
        self.coref_end_all_mlps = FullyConnectedLayer(config, config.hidden_size, self.all_cats_size, self.dropout_prob)

        self.antecedent_s2s_all_weights = nn.Parameter(torch.empty((self.num_cats, self.ffnn_size, self.ffnn_size)))
        self.antecedent_e2e_all_weights = nn.Parameter(torch.empty((self.num_cats, self.ffnn_size, self.ffnn_size)))
        self.antecedent_s2e_all_weights = nn.Parameter(torch.empty((self.num_cats, self.ffnn_size, self.ffnn_size)))
        self.antecedent_e2s_all_weights = nn.Parameter(torch.empty((self.num_cats, self.ffnn_size, self.ffnn_size)))

        self.antecedent_s2s_all_biases = nn.Parameter(torch.empty((self.num_cats, self.ffnn_size)))
        self.antecedent_e2e_all_biases = nn.Parameter(torch.empty((self.num_cats, self.ffnn_size)))
        self.antecedent_s2e_all_biases = nn.Parameter(torch.empty((self.num_cats, self.ffnn_size)))
        self.antecedent_e2s_all_biases = nn.Parameter(torch.empty((self.num_cats, self.ffnn_size)))

        self.reset_parameters()
        self.init_weights()

    def reset_parameters(self) -> None:
        W = [self.antecedent_s2s_all_weights, self.antecedent_e2e_all_weights,
             self.antecedent_s2e_all_weights, self.antecedent_e2s_all_weights]

        B = [self.antecedent_s2s_all_biases, self.antecedent_e2e_all_biases,
             self.antecedent_s2e_all_biases, self.antecedent_e2s_all_biases]

        for w, b in zip(W, B):
            init.kaiming_uniform_(w, a=math.sqrt(5))
            fan_in, _ = init._calculate_fan_in_and_fan_out(w)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(b, -bound, bound)

    def num_parameters(self) -> tuple:
        def head_filter(x):
            return x[1].requires_grad and any(hp in x[0] for hp in ['coref', 'mention', 'antecedent'])

        head_params = filter(head_filter, self.named_parameters())
        head_params = sum(p.numel() for n, p in head_params)
        return super().num_parameters() - head_params, head_params

    def _get_span_mask(self, batch_size, k, max_k):
        """
        :param batch_size: int
        :param k: tensor of size [batch_size], with the required k for each example
        :param max_k: int
        :return: [batch_size, max_k] of zero-ones, where 1 stands for a valid span and 0 for a padded span
        """
        size = (batch_size, max_k)
        idx = torch.arange(max_k, device=self.device).unsqueeze(0).expand(size)
        len_expanded = k.unsqueeze(1).expand(size)
        return (idx < len_expanded).int()

    def _prune_topk_mentions(self, mention_logits, attention_mask):
        """
        :param mention_logits: Shape [batch_size, seq_length, seq_length]
        :param attention_mask: [batch_size, seq_length]
        :param top_lambda:
        :return:
        """
        batch_size, seq_length, _ = mention_logits.size()
        actual_seq_lengths = torch.sum(attention_mask, dim=-1)  # [batch_size]

        k = (actual_seq_lengths * self.top_lambda).int()  # [batch_size]
        max_k = int(torch.max(k))  # This is the k for the largest input in the batch, we will need to pad

        _, topk_1d_indices = torch.topk(mention_logits.view(batch_size, -1), dim=-1, k=max_k)  # [batch_size, max_k]

        span_mask = self._get_span_mask(batch_size, k, max_k)  # [batch_size, max_k]
        # drop the invalid indices and set them to the last index
        topk_1d_indices = (topk_1d_indices * span_mask) + (1 - span_mask) * ((seq_length ** 2) - 1)  # We take different k for each example
        # sorting for coref mention order
        sorted_topk_1d_indices, _ = torch.sort(topk_1d_indices, dim=-1)  # [batch_size, max_k]

        # gives the row index in 2D matrix
        topk_mention_start_ids = torch.div(sorted_topk_1d_indices, seq_length, rounding_mode='floor') # [batch_size, max_k]
        topk_mention_end_ids = sorted_topk_1d_indices % seq_length  # [batch_size, max_k]

        topk_mention_logits = mention_logits[torch.arange(batch_size).unsqueeze(-1).expand(batch_size, max_k),
                                             topk_mention_start_ids, topk_mention_end_ids]  # [batch_size, max_k]

        # this is antecedents scores - rows mentions, cols coref mentions
        topk_mention_logits = topk_mention_logits.unsqueeze(-1) + topk_mention_logits.unsqueeze(-2)  # [batch_size, max_k, max_k]

        return topk_mention_start_ids, topk_mention_end_ids, span_mask, topk_mention_logits, topk_1d_indices

    def _mask_antecedent_logits(self, antecedent_logits, span_mask, categories_masks=None):
        antecedents_mask = torch.ones_like(antecedent_logits, dtype=self.dtype).tril(diagonal=-1)

        if categories_masks is not None:
            mask = antecedents_mask * span_mask.unsqueeze(1).unsqueeze(-1)
            mask *= categories_masks
        else:
            mask = antecedents_mask * span_mask.unsqueeze(-1)

        antecedent_logits = mask_tensor(antecedent_logits, mask)
        return antecedent_logits

    def _get_clusters_labels(self, span_starts, span_ends, all_clusters):
        """
        :param span_starts: [batch_size, max_k]
        :param span_ends: [batch_size, max_k]
        :param all_clusters: [batch_size, max_cluster_size, max_clusters_num, 2]
        :return: [batch_size, max_k, max_k + 1] - [b, i, j] == 1 if j is antecedent of i
        """
        batch_size, max_k = span_starts.size()
        new_cluster_labels = np.zeros((batch_size, max_k, max_k))

        span_starts_cpu = span_starts.cpu().tolist()
        span_ends_cpu = span_ends.cpu().tolist()
        all_clusters_cpu = all_clusters.cpu().tolist()

        for b, (starts, ends, gold_clusters) in enumerate(zip(span_starts_cpu, span_ends_cpu, all_clusters_cpu)):
            gold_clusters = extract_clusters(gold_clusters)
            mention_to_gold_clusters = extract_mentions_to_clusters(gold_clusters)
            for i, (start, end) in enumerate(zip(starts, ends)):
                if (start, end) not in mention_to_gold_clusters:
                    continue
                for j, (a_start, a_end) in enumerate(list(zip(starts, ends))[:i]):
                    if (a_start, a_end) in mention_to_gold_clusters[(start, end)]:
                        new_cluster_labels[b, i, j] = 1

        new_cluster_labels = torch.tensor(new_cluster_labels, device=self.device)
        return new_cluster_labels

    def _get_categories_labels(self, tokens, subtoken_map, new_token_map, span_starts, span_ends):
        batch_size, max_k = span_starts.size()

        spans = []
        for b, (starts, ends) in enumerate(zip(span_starts.cpu().tolist(), span_ends.cpu().tolist())):
            doc_spans = []
            for start, end in zip(starts, ends):
                token_indices = [new_token_map[b][idx] for idx in set(subtoken_map[b][start:end + 1]) - {None}]
                span = {tokens[b][idx].lower() for idx in token_indices if idx is not None}
                pronoun_id = get_pronoun_id(span)
                doc_spans.append((span - STOPWORDS, pronoun_id))
            spans.append(doc_spans)

        categories_labels = np.zeros((batch_size, max_k, max_k)) - 1
        for b in range(batch_size):
            for i in range(max_k):
                for j in list(range(max_k))[:i]:
                    categories_labels[b, i, j] = get_category_id(spans[b][i], spans[b][j])

        categories_labels = torch.tensor(categories_labels, device=self.device)
        categories_masks = [categories_labels == cat_id for cat_id in range(self.num_cats - 1)] + [categories_labels != -1]
        categories_masks = torch.stack(categories_masks, dim=1).int()
        return categories_labels, categories_masks

    def _get_marginal_log_likelihood_loss(self, logits, labels, span_mask):
        gold_coref_logits = mask_tensor(logits, labels)                       # [batch_size, num_cats + 1, max_k, max_k]

        gold_log_sum_exp = torch.logsumexp(gold_coref_logits, dim=-1)         # [batch_size, num_cats + 1, max_k]
        all_log_sum_exp = torch.logsumexp(logits, dim=-1)                     # [batch_size, num_cats + 1, max_k]
        losses = all_log_sum_exp - gold_log_sum_exp                           # [batch_size, num_cats + 1, max_k]

        # zero the loss of padded spans
        span_mask = span_mask.unsqueeze(1)                                    # [batch_size, 1, max_k]
        losses = losses * span_mask                                           # [batch_size, num_cats, max_k]

        # normalize loss by spans
        per_span_loss = losses.mean(dim=-1)                                   # [batch_size, num_cats + 1]

        # normalize loss by document
        loss_per_cat = per_span_loss.mean(dim=0)                              # [num_cats + 1]

        # normalize loss by category
        loss = loss_per_cat.sum()
        return loss

    def _get_mention_mask(self, mention_logits_or_weights):
        """
        Returns a tensor of size [batch_size, seq_length, seq_length] where valid spans
        (start <= end < start + max_span_length) are 1 and the rest are 0
        :param mention_logits_or_weights: Either the span mention logits or weights, size [batch_size, seq_length, seq_length]
        """
        mention_mask = torch.ones_like(mention_logits_or_weights, dtype=self.dtype)
        mention_mask = mention_mask.triu(diagonal=0)
        mention_mask = mention_mask.tril(diagonal=self.max_span_length - 1)
        return mention_mask

    def _calc_mention_logits(self, start_mention_reps, end_mention_reps):
        start_mention_logits = self.mention_start_classifier(start_mention_reps).squeeze(-1)  # [batch_size, seq_length]
        end_mention_logits = self.mention_end_classifier(end_mention_reps).squeeze(-1)        # [batch_size, seq_length]

        temp = self.mention_s2e_classifier(start_mention_reps)                                # [batch_size, seq_length]
        joint_mention_logits = torch.matmul(temp,
                                            end_mention_reps.permute([0, 2, 1]))              # [batch_size, seq_length, seq_length]

        mention_logits = joint_mention_logits + start_mention_logits.unsqueeze(-1) + end_mention_logits.unsqueeze(-2)
        mention_mask = self._get_mention_mask(mention_logits)                                 # [batch_size, seq_length, seq_length]
        mention_logits = mask_tensor(mention_logits, mention_mask)                            # [batch_size, seq_length, seq_length]
        return mention_logits

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_cats, self.ffnn_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)    # bnkf/bnlg

    def _calc_coref_logits(self, start_reps, end_reps):
        # see discussion on einsum: https://discuss.pytorch.org/t/batch-matrix-multiplication-of-3d-tensors/153644/4

        all_starts = self.transpose_for_scores(self.coref_start_all_mlps(start_reps))
        all_ends = self.transpose_for_scores(self.coref_end_all_mlps(end_reps))

        logits = torch.einsum('bnkf, nfg, bnlg -> bnkl', all_starts, self.antecedent_s2s_all_weights, all_starts) + \
                 torch.einsum('bnkf, nfg, bnlg -> bnkl', all_ends,   self.antecedent_e2e_all_weights, all_ends) + \
                 torch.einsum('bnkf, nfg, bnlg -> bnkl', all_starts, self.antecedent_s2e_all_weights, all_ends) + \
                 torch.einsum('bnkf, nfg, bnlg -> bnkl', all_ends,   self.antecedent_e2s_all_weights, all_starts)

        biases = torch.einsum('bnkf, nf -> bnk', all_starts, self.antecedent_s2s_all_biases).unsqueeze(-2) + \
                 torch.einsum('bnkf, nf -> bnk', all_ends,   self.antecedent_e2e_all_biases).unsqueeze(-2) + \
                 torch.einsum('bnkf, nf -> bnk', all_ends,   self.antecedent_s2e_all_biases).unsqueeze(-2) + \
                 torch.einsum('bnkf, nf -> bnk', all_starts, self.antecedent_e2s_all_biases).unsqueeze(-2)

        return logits + biases

    def _get_all_labels(self, clusters_labels, categories_masks):
        batch_size, max_k, _ = clusters_labels.size()

        categories_labels = clusters_labels.unsqueeze(1).repeat(1, self.num_cats, 1, 1) * categories_masks
        all_labels = torch.cat((categories_labels, clusters_labels.unsqueeze(1)), dim=1)            # for the combined loss (L_coref + L_tasks)

        # null cluster
        zeros = torch.zeros((batch_size, self.num_cats + 1, max_k, 1), device=self.device)
        all_labels = torch.cat((all_labels, zeros), dim=-1)                                         # [batch_size, num_cats + 1, max_k, max_k + 1]
        no_antecedents = 1 - torch.sum(all_labels, dim=-1).bool().float()
        all_labels[:, :, :, -1] = no_antecedents

        return all_labels

    def forward_transformer(self, batch):
        input_ids = batch['input_ids']
        attention_mask = batch['attention_mask']

        if 'leftovers' not in batch:
            outputs = self.base_model(input_ids, attention_mask=attention_mask)
            sequence_output = outputs.last_hidden_state
        else:
            docs, segments, segment_len = input_ids.size()
            input_ids, attention_mask = input_ids.view(-1, segment_len), attention_mask.view(-1, segment_len)

            outputs = self.base_model(input_ids, attention_mask=attention_mask)
            sequence_output = outputs.last_hidden_state

            attention_mask = attention_mask.view((docs, segments * segment_len))        # [docs, seq_len]
            sequence_output = sequence_output.view((docs, segments * segment_len, -1))  # [docs, seq_len, dim]

            leftovers_ids, leftovers_mask = batch['leftovers']['input_ids'], batch['leftovers']['attention_mask']
            if len(leftovers_ids) > 0:
                res_outputs = self.base_model(leftovers_ids, attention_mask=leftovers_mask)
                res_sequence_output = res_outputs.last_hidden_state

                attention_mask = torch.cat([attention_mask, leftovers_mask], dim=1)
                sequence_output = torch.cat([sequence_output, res_sequence_output], dim=1)

        return sequence_output, attention_mask

    def forward(self, batch, gold_clusters=None, return_all_outputs=False):
        tokens, subtoken_map, new_token_map = batch['tokens'], batch['subtoken_map'], batch['new_token_map']

        sequence_output, attention_mask = self.forward_transformer(batch)

        # Compute representations
        start_mention_reps = self.start_mention_mlp(sequence_output)
        end_mention_reps = self.end_mention_mlp(sequence_output)

        # mention scores
        mention_logits = self._calc_mention_logits(start_mention_reps, end_mention_reps)

        # prune mentions
        mention_start_ids, mention_end_ids, span_mask, topk_mention_logits, topk_1d_indices = self._prune_topk_mentions(mention_logits, attention_mask)

        categories_labels, categories_masks = self._get_categories_labels(
            tokens, subtoken_map, new_token_map, mention_start_ids, mention_end_ids
        )

        batch_size, max_k = mention_start_ids.size()
        size = (batch_size, max_k, self.hidden_size)
        # gather reps
        topk_start_reps = torch.gather(sequence_output, dim=1, index=mention_start_ids.unsqueeze(-1).expand(size))
        topk_end_reps = torch.gather(sequence_output, dim=1, index=mention_end_ids.unsqueeze(-1).expand(size))

        # antecedent scores by category
        categories_logits = self._calc_coref_logits(topk_start_reps, topk_end_reps)

        final_logits = categories_logits * categories_masks
        final_logits = final_logits.sum(dim=1) + topk_mention_logits
        categories_logits = categories_logits + topk_mention_logits.unsqueeze(1)

        # lower logits of padded spans or different category.
        final_logits = self._mask_antecedent_logits(final_logits, span_mask)
        categories_logits = self._mask_antecedent_logits(categories_logits, span_mask, categories_masks)

        # adding zero logits for null span
        final_logits = torch.cat((final_logits, torch.zeros((batch_size, max_k, 1), device=self.device)), dim=-1)                           # [batch_size, max_k, max_k + 1]
        categories_logits = torch.cat((categories_logits, torch.zeros((batch_size, self.num_cats, max_k, 1), device=self.device)), dim=-1)  # [batch_size, num_cats, max_k, max_k + 1]

        if return_all_outputs:
            outputs = (mention_start_ids, mention_end_ids, mention_logits, final_logits)
        else:
            outputs = tuple()

        if gold_clusters is not None:
            clusters_labels = self._get_clusters_labels(mention_start_ids, mention_end_ids, gold_clusters)
            all_labels = self._get_all_labels(clusters_labels, categories_masks)
            all_logits = torch.cat((categories_logits, final_logits.unsqueeze(1)), dim=1)

            loss = self._get_marginal_log_likelihood_loss(all_logits, all_labels, span_mask)
            outputs = (loss,) + outputs + (categories_labels, clusters_labels, )

        return outputs




from spacy import Language, util
from spacy.tokens import Doc, Span
from fastcoref import FCoref, LingMessCoref
from typing import List,Tuple


@Language.factory(
    "fastcoref",
    assigns=["doc._.resolved_text","doc._.coref_clusters"],
    default_config={
        "model_architecture": 'FCoref', # FCoref or LingMessCoref 
        "model_path": 'biu-nlp/f-coref', # You can specify your own trained model path
        "device": None, # "cuda" or "cpu" None defaults to cuda
        "max_tokens_in_batch":10000
    },
)
class FastCorefResolver:
    """a class that implements the logic from
    https://towardsdatascience.com/how-to-make-an-effective-coreference-resolution-model-55875d2b5f19"""
    def __init__(
        self,
        nlp,
        name,
        model_architecture: str,
        model_path: str,
        device,
        max_tokens_in_batch: int
    ):
        assert model_architecture in ["FCoref","LingMessCoref"]
        if model_architecture == "FCoref": 
            self.coref_model = FCoref(model_name_or_path = model_path, device=device, nlp=nlp)
        elif model_architecture == "LingMessCoref":
            self.coref_model = LingMessCoref(model_name_or_path = model_path, device=device, nlp=nlp)
        self.max_tokens_in_batch = max_tokens_in_batch
        # Register custom extension on the Doc
        if not Doc.has_extension("resolved_text"):
            Doc.set_extension("resolved_text", default="")
        if not Doc.has_extension("coref_clusters"):
            Doc.set_extension("coref_clusters", default=None)
         
    def _get_span_noun_indices(self, doc: Doc, cluster: List[Tuple]) -> List[int]:
        """
        > Get the indices of the spans in the cluster that contain at least one noun or proper noun
        :param doc: Doc
        :param cluster: List[Tuple]
        :return: A list of indices of spans that contain at least one noun or proper noun.
        """
        spans = [doc.char_span(span[0],span[1]) for span in cluster]
        spans_pos = [[token.pos_ for token in span] for span in spans]
        span_noun_indices = [
            i for i, span_pos in enumerate(spans_pos) if any(pos in span_pos for pos in ["NOUN", "PROPN"])
        ]
        return span_noun_indices
    def _get_cluster_head(self, doc: Doc, cluster: List[Tuple], noun_indices: List[int]):
        """
        > Given a spaCy Doc, a coreference cluster, and a list of noun indices, return the head span and its start and end
        indices
        :param doc: the spaCy Doc object
        :type doc: Doc
        :param cluster: a list of Tuples, where each tuple is a char indices of token in the document
        :type cluster: List[Tuple]
        :param noun_indices: a list of indices of the nouns in the cluster
        :type noun_indices: List[int]
        :return: The head span and the start and end indices of the head span.
        """
        head_idx = noun_indices[0]
        head_start,head_end = cluster[head_idx]
        head_span = doc.char_span(head_start,head_end)
        return head_span, [head_start, head_end]
    def _is_containing_other_spans(self,span: List[int], all_spans: List[List[int]]):
        """
        It returns True if there is any span in all_spans that is contained within span and is not equal to span
        :param span: the span we're checking to see if it contains other spans
        :type span: List[int]
        :param all_spans: a list of all the spans in the document
        :type all_spans: List[List[int]]
        :return: A list of all spans that are not contained in any other span.
        """
        return any([s[0] >= span[0] and s[1] <= span[1] and s != span for s in all_spans])
    def _core_logic_part(self,document: Doc, coref: List[int], resolved: List[str], mention_span: Span):
        """
        If the last token of the mention is a possessive pronoun, then add an apostrophe and an s to the mention.
        Otherwise, just add the last token to the mention
        :param document: Doc object
        :type document: Doc
        :param coref: List[int]
        :param resolved: list of strings, where each string is a token in the sentence
        :param mention_span: The span of the mention that we want to replace
        :return: The resolved list is being returned.
        """
        char_span = document.char_span(coref[0],coref[1])
        final_token = char_span[-1]
        final_token_tag = str(final_token.tag_).lower()
        test_token_test = False
        for option in ["PRP$", "POS", "BEZ"]:
            if option.lower() in final_token_tag:
                test_token_test = True
                break
        if test_token_test:
            resolved[char_span.start] = mention_span.text + "'s" + final_token.whitespace_
        else:
            resolved[char_span.start] = mention_span.text + final_token.whitespace_
        for i in range(char_span.start + 1, char_span.end):
            resolved[i] = ""
        return resolved

    def __call__(self, doc: Doc, resolve_text=False) -> Doc:
        """
        The function takes a doc object and returns a doc object
        :param doc: Doc
        :type doc: Doc
        :return: A Doc object with the resolved text and coreference clusters added as attributes.
        """
        preds = self.coref_model.predict(texts=[doc.text])
        clusters = preds[0].get_clusters(as_strings=False)
        if resolve_text:
            resolved = list(tok.text_with_ws for tok in doc)
            all_spans = [span for cluster in clusters for span in cluster]
            for cluster in clusters:
                indices = self._get_span_noun_indices(doc,cluster)
                if indices:
                    mention_span, mention = self._get_cluster_head(doc, cluster, indices)
                    for coref in cluster:
                        if coref != mention and not self._is_containing_other_spans(coref, all_spans):
                            self._core_logic_part(doc, coref, resolved, mention_span)
            doc._.resolved_text = "".join(resolved)
        doc._.coref_clusters = clusters
        return doc

    def pipe(self, stream, batch_size=512, resolve_text=False):
        for docs in util.minibatch(stream, size=batch_size):
            preds = self.coref_model.predict(
                    texts=[doc.text for doc in docs],max_tokens_in_batch=self.max_tokens_in_batch)
            for idx,pred in enumerate(preds):
                clusters = pred.get_clusters(as_strings=False)
                doc = docs[idx] 
                if resolve_text:    
                    resolved = list(tok.text_with_ws for tok in doc)
                    all_spans = [span for cluster in clusters for span in cluster]
                    for cluster in clusters:
                            indices = self._get_span_noun_indices(doc,cluster)
                            if indices:
                                mention_span, mention = self._get_cluster_head(doc, cluster, indices)
                                for coref in cluster:
                                    if coref != mention and not self._is_containing_other_spans(coref, all_spans):
                                        self._core_logic_part(doc, coref, resolved, mention_span)
                    doc._.resolved_text = "".join(resolved)
                doc._.coref_clusters = clusters
                yield doc


SUPPORTED_MODELS = ['bert', 'roberta', 'distilbert', 'distilroberta', 'longformer']
SPEAKER_START = '#####'
SPEAKER_END = '###'
NULL_ID_FOR_COREF = -1


PRONOUNS_GROUPS = {
            'i': 0, 'me': 0, 'my': 0, 'mine': 0, 'myself': 0,
            'you': 1, 'your': 1, 'yours': 1, 'yourself': 1, 'yourselves': 1,
            'he': 2, 'him': 2, 'his': 2, 'himself': 2,
            'she': 3, 'her': 3, 'hers': 3, 'herself': 3,
            'it': 4, 'its': 4, 'itself': 4,
            'we': 5, 'us': 5, 'our': 5, 'ours': 5, 'ourselves': 5,
            'they': 6, 'them': 6, 'their': 6, 'themselves': 6,
            'that': 7, 'this': 7
}

STOPWORDS = {"'s", 'a', 'all', 'an', 'and', 'at', 'for', 'from', 'in', 'into',
             'more', 'of', 'on', 'or', 'some', 'the', 'these', 'those'}

CATEGORIES = {'pron-pron-comp': 0,
              'pron-pron-no-comp': 1,
              'pron-ent': 2,
              'match': 3,
              'contain': 4,
              'other': 5
              }




import os
import logging
import numpy as np
import torch
import spacy
import wandb
from dataclasses import dataclass

from torch.optim.adamw import AdamW
from tqdm.auto import tqdm

import transformers
from transformers import AutoConfig, AutoTokenizer, get_linear_schedule_with_warmup

from fastcoref.coref_models.modeling_fcoref import FCorefModel

from fastcoref.utilities import coref_dataset
from fastcoref.utilities.collate import DynamicBatchSampler, LeftOversCollator
from fastcoref.utilities.consts import SUPPORTED_MODELS
from fastcoref.utilities.metrics import MentionEvaluator, CorefEvaluator
from fastcoref.utilities.util import set_seed, create_mention_to_antecedent, create_clusters, \
    output_evaluation_metrics, update_metrics, save_all

# Setup logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s - %(levelname)s - \t %(message)s',
                    datefmt='%m/%d/%Y %H:%M:%S', level=logging.INFO)


@dataclass
class TrainingArgs:
    model_name_or_path: str
    output_dir: str
    overwrite_output_dir: bool = False
    learning_rate: float = 1e-5
    head_learning_rate: float = 3e-4
    dropout_prob: float = 0.3
    weight_decay: float = 0.01
    adam_beta1: float = 0.9
    adam_beta2: float = 0.98
    adam_epsilon: float = 1e-6
    epochs: float = 3
    ffnn_size: int = 1024
    logging_steps: int = 500
    eval_steps: int = 500
    seed: int = 42
    max_span_length: int = 30
    top_lambda: float = 0.4
    cache_dir: str = 'cache'
    max_segment_len: int = 512
    max_doc_len: int = None
    max_tokens_in_batch: int = 5000
    device: str = None


def _load_f_coref_model(args):
    logger.info(f'Loading FCoref model with underlying transformer {args.model_name_or_path}')

    config = AutoConfig.from_pretrained(args.model_name_or_path, cache_dir=args.cache_dir)
    config.coref_head = {
        "max_span_length": args.max_span_length,
        "top_lambda": args.top_lambda,
        "ffnn_size": args.ffnn_size,
        "dropout_prob": args.dropout_prob,
        "max_segment_len": args.max_segment_len
    }

    tokenizer = AutoTokenizer.from_pretrained(
        args.model_name_or_path, use_fast=True, add_prefix_space=True, cache_dir=args.cache_dir
    )

    model = FCorefModel.from_pretrained(
        args.model_name_or_path, output_loading_info=False,
        config=config, cache_dir=args.cache_dir
    )

    t_params, h_params = [p / 1000000 for p in model.num_parameters()]
    logger.info(f'FCoref Parameters: {t_params + h_params:.1f}M, '
                f'Transformer: {t_params:.1f}M, Coref head: {h_params:.1f}M')

    if model.base_model_prefix not in SUPPORTED_MODELS:
        raise NotImplementedError(f'Not supporting {model.base_model_prefix}, choose one of {SUPPORTED_MODELS}')

    return model, tokenizer


class CorefTrainer:
    def __init__(self, args: TrainingArgs, train_file, dev_file=None, test_file=None):
        transformers.logging.set_verbosity_error()
        self.args = args
        wandb.init(project=self.args.output_dir, config=self.args)

        self._set_device()

        self.nlp = spacy.load("en_core_web_sm", exclude=["tagger", "parser", "lemmatizer", "ner", "textcat"])
        self.model, self.tokenizer = _load_f_coref_model(self.args)
        self.model.to(self.device)

        self.collator = LeftOversCollator(
            tokenizer=self.tokenizer,
            device=self.device,
            max_segment_len=self.args.max_segment_len
        )

        self.train_dataset, self.train_sampler = self._get_sampler(train_file)

        self.dev_dataset, self.dev_sampler = None, None
        if dev_file is not None:
            self.dev_dataset, self.dev_sampler = self._get_sampler(dev_file)

        self.test_dataset, self.test_sampler = None, None
        if test_file is not None:
            self.test_dataset, self.test_sampler = self._get_sampler(test_file)

        set_seed(self.args)

    def _get_sampler(self, file):
        dataset = coref_dataset.create(
            file=file, tokenizer=self.tokenizer, nlp=self.nlp
        )
        sampler = DynamicBatchSampler(
            dataset=dataset,
            collator=self.collator,
            max_tokens=self.args.max_tokens_in_batch,
            max_segment_len=self.args.max_segment_len
        )

        return dataset, sampler

    def _set_device(self):
        # Setup CUDA, GPU & distributed training
        if self.args.device is None:
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
            self.device = torch.device(self.args.device)
        # TODO: this is not true
        self.args.n_gpu = torch.cuda.device_count()

    def train(self):
        """ Train the model """

        # we create batches beacuse the sampler is generating batches of sorted docs -> to avoid many pad tokens
        # so, we need to shuffle the batches somehow.
        train_batches = coref_dataset.create_batches(self.train_sampler, shuffle=True)

        t_total = len(train_batches) * self.args.epochs

        # Prepare optimizer and schedule (linear warmup and decay)
        no_decay = ['bias', 'LayerNorm.weight']
        head_params = ['coref', 'mention', 'antecedent']

        model_decay = [p for n, p in self.model.named_parameters() if
                       not any(hp in n for hp in head_params) and not any(nd in n for nd in no_decay)]
        model_no_decay = [p for n, p in self.model.named_parameters() if
                          not any(hp in n for hp in head_params) and any(nd in n for nd in no_decay)]
        head_decay = [p for n, p in self.model.named_parameters() if
                      any(hp in n for hp in head_params) and not any(nd in n for nd in no_decay)]
        head_no_decay = [p for n, p in self.model.named_parameters() if
                         any(hp in n for hp in head_params) and any(nd in n for nd in no_decay)]

        head_learning_rate = self.args.head_learning_rate if self.args.head_learning_rate else self.args.learning_rate
        optimizer_grouped_parameters = [
            {'params': model_decay, 'lr': self.args.learning_rate, 'weight_decay': self.args.weight_decay},
            {'params': model_no_decay, 'lr': self.args.learning_rate, 'weight_decay': 0.0},
            {'params': head_decay, 'lr': head_learning_rate, 'weight_decay': self.args.weight_decay},
            {'params': head_no_decay, 'lr': head_learning_rate, 'weight_decay': 0.0}
        ]
        optimizer = AdamW(optimizer_grouped_parameters,
                          lr=self.args.learning_rate,
                          betas=(self.args.adam_beta1, self.args.adam_beta2),
                          eps=self.args.adam_epsilon)
        scheduler = get_linear_schedule_with_warmup(optimizer, num_warmup_steps=t_total * 0.1,
                                                    num_training_steps=t_total)

        # using mixed precision
        scaler = torch.cuda.amp.GradScaler()

        # Train!
        logger.info("***** Running training *****")
        logger.info("  Num Epochs = %d", self.args.epochs)
        logger.info("  Total optimization steps = %d", t_total)

        global_step, tr_loss, logging_loss = 0, 0.0, 0.0
        best_f1, best_global_step = -1, -1

        train_iterator = tqdm(range(int(self.args.epochs)), desc="Epoch")
        for _ in train_iterator:
            epoch_iterator = tqdm(train_batches, desc="Iteration")
            for step, batch in enumerate(epoch_iterator):
                batch['input_ids'] = torch.tensor(batch['input_ids'], device=self.device)
                batch['attention_mask'] = torch.tensor(batch['attention_mask'], device=self.device)
                batch['gold_clusters'] = torch.tensor(batch['gold_clusters'], device=self.device)
                if 'leftovers' in batch:
                    batch['leftovers']['input_ids'] = torch.tensor(batch['leftovers']['input_ids'], device=self.device)
                    batch['leftovers']['attention_mask'] = torch.tensor(batch['leftovers']['attention_mask'],
                                                                        device=self.device)

                self.model.zero_grad()
                self.model.train()

                with torch.cuda.amp.autocast():
                    outputs = self.model(batch, gold_clusters=batch['gold_clusters'], return_all_outputs=False)

                loss = outputs[0]  # model outputs are always tuple in transformers (see doc)

                tr_loss += loss.item()
                scaler.scale(loss).backward()

                scaler.step(optimizer)
                scheduler.step()  # Update learning rate schedule
                scaler.update()  # Updates the scale for next iteration
                global_step += 1

                # Log metrics
                if global_step % self.args.logging_steps == 0:
                    loss = (tr_loss - logging_loss) / self.args.logging_steps
                    logger.info(f"loss step {global_step}: {loss}")
                    wandb.log({'loss': loss}, step=global_step)
                    logging_loss = tr_loss

                # Evaluation
                if self.dev_sampler is not None and global_step % self.args.eval_steps == 0:
                    results = self.evaluate(prefix=f'step_{global_step}', test=False)
                    wandb.log(results, step=global_step)

                    f1 = results["f1"]
                    if f1 > best_f1:
                        best_f1, best_global_step = f1, global_step
                        wandb.run.summary["best_f1"] = best_f1

                        # Save model
                        output_dir = os.path.join(self.args.output_dir, f'model')
                        save_all(tokenizer=self.tokenizer, model=self.model, output_dir=output_dir)
                    logger.info(f"best f1 is {best_f1} on global step {best_global_step}")

    def evaluate(self, test=False, prefix=''):
        if test:
            eval_sampler = self.test_sampler
            dataset_str = 'test_set'
        else:
            eval_sampler = self.dev_sampler
            dataset_str = 'dev_set'
        if eval_sampler is None:
            logger.info(f'Skipping evaluation. {dataset_str} is None')
            return {}

        self.model.eval()

        logger.info(f"***** Running evaluation on {dataset_str} - {len(eval_sampler.dataset)} documents *****")

        metrics_dict = {'loss': 0., 'post_pruning': MentionEvaluator(), 'mentions': MentionEvaluator(),
                        'coref': CorefEvaluator()}
        doc_to_tokens = {}
        doc_to_subtoken_map = {}
        doc_to_new_word_map = {}
        doc_to_prediction = {}

        with tqdm(desc="Inference", total=len(eval_sampler.dataset)) as progress_bar:
            for idx, batch in enumerate(eval_sampler):
                doc_keys = batch['doc_key']
                tokens = batch['tokens']
                subtoken_map = batch['subtoken_map']
                new_token_map = batch['new_token_map']
                gold_clusters = batch['gold_clusters']

                with torch.no_grad():
                    outputs = self.model(batch, gold_clusters=gold_clusters, return_all_outputs=True)

                outputs_np = tuple(tensor.cpu().numpy() for tensor in outputs)

                gold_clusters = gold_clusters.cpu().numpy()
                loss, span_starts, span_ends, mention_logits, coref_logits = outputs_np
                metrics_dict['loss'] += loss.item()

                doc_indices, mention_to_antecedent = create_mention_to_antecedent(span_starts, span_ends, coref_logits)

                for i, doc_key in enumerate(doc_keys):
                    doc_mention_to_antecedent = mention_to_antecedent[np.nonzero(doc_indices == i)]
                    predicted_clusters = create_clusters(doc_mention_to_antecedent)

                    doc_to_prediction[doc_key] = predicted_clusters
                    doc_to_tokens[doc_key] = tokens[i]
                    doc_to_subtoken_map[doc_key] = subtoken_map[i]
                    doc_to_new_word_map[doc_key] = new_token_map[i]

                    update_metrics(metrics_dict, span_starts[i], span_ends[i], gold_clusters[i], predicted_clusters)

                progress_bar.update(n=len(doc_keys))

        results = output_evaluation_metrics(
            metrics_dict=metrics_dict, prefix=prefix
        )

        return results

    def push_to_hub(self, repo_name, organization=None):
        self.model.push_to_hub(repo_name, organization=organization, use_temp_dir=True)
        self.tokenizer.push_to_hub(repo_name, organization=organization, use_temp_dir=True)



import logging
import os
import random
import torch
import numpy as np
from fastcoref.utilities.consts import NULL_ID_FOR_COREF, CATEGORIES, PRONOUNS_GROUPS

logger = logging.getLogger(__name__)


def flatten(l):
    return [item for sublist in l for item in sublist]


def save_all(model, tokenizer, output_dir):
    logger.info(f"Saving model to {output_dir}")
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Save a trained model, configuration and tokenizer using `save_pretrained()`.
    # They can then be reloaded using `from_pretrained()`
    model.save_pretrained(output_dir)
    tokenizer.save_pretrained(output_dir)


def pad_clusters_inside(clusters, max_cluster_size):
    return [cluster + [(NULL_ID_FOR_COREF, NULL_ID_FOR_COREF)] * (max_cluster_size - len(cluster)) for cluster
            in clusters]


def pad_clusters_outside(clusters, max_num_clusters):
    return clusters + [[]] * (max_num_clusters - len(clusters))


def pad_clusters(clusters, max_num_clusters, max_cluster_size):
    clusters = pad_clusters_outside(clusters, max_num_clusters)
    clusters = pad_clusters_inside(clusters, max_cluster_size)
    return clusters


def output_evaluation_metrics(metrics_dict, prefix):
    loss = metrics_dict['loss']
    post_pruning_mention_pr, post_pruning_mentions_r, post_pruning_mention_f1 = metrics_dict['post_pruning'].get_prf()
    mention_p, mentions_r, mention_f1 = metrics_dict['mentions'].get_prf()
    p, r, f1 = metrics_dict['coref'].get_prf()
    results = {
        'eval_loss': loss,
        "post pruning mention precision": post_pruning_mention_pr,
        "post pruning mention recall": post_pruning_mentions_r,
        "post pruning mention f1": post_pruning_mention_f1,
        "mention precision": mention_p,
        "mention recall": mentions_r,
        "mention f1": mention_f1,
        "precision": p,
        "recall": r,
        "f1": f1
    }

    logger.info("***** Eval results {} *****".format(prefix))
    for key, value in results.items():
        if isinstance(value, float):
            logger.info(f"  {key : <30} = {value:.3f}")
        elif isinstance(value, dict):
            logger.info(f"  {key : <30} = {value}")

    return results



def update_metrics(metrics, span_starts, span_ends, gold_clusters, predicted_clusters):
    gold_clusters = extract_clusters(gold_clusters)
    candidate_mentions = list(zip(span_starts, span_ends))

    mention_to_gold_clusters = extract_mentions_to_clusters(gold_clusters)
    mention_to_predicted_clusters = extract_mentions_to_clusters(predicted_clusters)

    gold_mentions = list(mention_to_gold_clusters.keys())
    predicted_mentions = list(mention_to_predicted_clusters.keys())

    metrics['post_pruning'].update(candidate_mentions, gold_mentions)
    metrics['mentions'].update(predicted_mentions, gold_mentions)
    metrics['coref'].update(predicted_clusters, gold_clusters,
                            mention_to_predicted_clusters, mention_to_gold_clusters)


def encode(batch, tokenizer, nlp):
    tokenized_texts = tokenize_with_spacy(batch['text'], nlp)
    encoded_batch = tokenizer(
        tokenized_texts['tokens'], add_special_tokens=True, is_split_into_words=True,
        return_length=True, return_attention_mask=False
    )
    return {
        'tokens': tokenized_texts['tokens'],
        'input_ids': encoded_batch['input_ids'],
        'length': encoded_batch['length'],

        # bpe token -> spacy tokens
        'subtoken_map': [enc.word_ids for enc in encoded_batch.encodings],
        # this is a can use for speaker info TODO: better name!
        'new_token_map': [list(range(len(tokens))) for tokens in tokenized_texts['tokens']],
        # spacy tokens -> text char
        'offset_mapping': tokenized_texts['offset_mapping']
    }


def tokenize_with_spacy(texts, nlp):
    def handle_doc(doc):
        tokens = []
        offset_mapping = []
        for tok in doc:
            tokens.append(tok.text)
            offset_mapping.append((tok.idx, tok.idx + len(tok.text)))
        return tokens, offset_mapping

    tokenized_texts = {'tokens': [], 'offset_mapping': []}

    # Edge case - Also disable other custom components
    all_pipe_names = nlp.pipe_names
    tokenizer_pipe_names = ["tok2vec"]

    disabled_pipe_names = [pipe_name for pipe_name in all_pipe_names if pipe_name not in tokenizer_pipe_names]
    docs = nlp.pipe(texts, disable=disabled_pipe_names)
    for doc in docs:
        tokens, offset_mapping = handle_doc(doc)
        tokenized_texts['tokens'].append(tokens)
        tokenized_texts['offset_mapping'].append(offset_mapping)

    return tokenized_texts


def align_to_char_level(span_starts, span_ends, token_to_char, subtoken_map=None, new_token_map=None):
    char_map = {}
    reverse_char_map = {}
    for idx, (start, end) in enumerate(zip(span_starts, span_ends)):
        new_start, new_end = start.copy(), end.copy()

        try:
            if subtoken_map is not None:
                new_start, new_end = subtoken_map[new_start], subtoken_map[new_end]
                if new_start is None or new_end is None:
                    char_map[(start, end)] = None, None
                    continue
            if new_token_map is not None:
                new_start, new_end = new_token_map[new_start], new_token_map[new_end]
            new_start, new_end = token_to_char[new_start][0], token_to_char[new_end][1]
            char_map[(start, end)] = idx, (new_start, new_end)
            reverse_char_map[(new_start, new_end)] = idx, (start, end)
        except IndexError:
            # this is padding index
            char_map[(start, end)] = None, None
            continue

    return char_map, reverse_char_map


def set_seed(args):
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    if args.n_gpu > 0:
        torch.cuda.manual_seed_all(args.seed)


def extract_clusters(gold_clusters):
    gold_clusters = [tuple(tuple(m) for m in cluster if NULL_ID_FOR_COREF not in m) for cluster in gold_clusters]
    gold_clusters = [cluster for cluster in gold_clusters if len(cluster) > 0]
    return gold_clusters


def extract_mentions_to_clusters(gold_clusters):
    mention_to_gold = {}
    for gc in gold_clusters:
        for mention in gc:
            mention_to_gold[mention] = gc
    return mention_to_gold


def create_clusters(mention_to_antecedent):
    # Note: mention_to_antecedent is a numpy array

    clusters, mention_to_cluster = [], {}
    for mention, antecedent in mention_to_antecedent:
        mention, antecedent = tuple(mention), tuple(antecedent)
        if antecedent in mention_to_cluster:
            cluster_idx = mention_to_cluster[antecedent]
            if mention not in clusters[cluster_idx]:
                clusters[cluster_idx].append(mention)
                mention_to_cluster[mention] = cluster_idx
        elif mention in mention_to_cluster:
            cluster_idx = mention_to_cluster[mention]
            if antecedent not in clusters[cluster_idx]:
                clusters[cluster_idx].append(antecedent)
                mention_to_cluster[antecedent] = cluster_idx
        else:
            cluster_idx = len(clusters)
            mention_to_cluster[mention] = cluster_idx
            mention_to_cluster[antecedent] = cluster_idx
            clusters.append([antecedent, mention])

    clusters = [tuple(cluster) for cluster in clusters]
    return clusters


def create_mention_to_antecedent(span_starts, span_ends, coref_logits):
    batch_size, n_spans, _ = coref_logits.shape

    max_antecedents = coref_logits.argmax(axis=-1)
    doc_indices, mention_indices = np.nonzero(max_antecedents < n_spans)        # indices where antecedent is not null.
    antecedent_indices = max_antecedents[max_antecedents < n_spans]
    span_indices = np.stack([span_starts, span_ends], axis=-1)

    mentions = span_indices[doc_indices, mention_indices]
    antecedents = span_indices[doc_indices, antecedent_indices]
    mention_to_antecedent = np.stack([mentions, antecedents], axis=1)

    return doc_indices, mention_to_antecedent


def mask_tensor(t, mask):
    t = t + ((1.0 - mask.float()) * -10000.0)
    t = torch.clamp(t, min=-10000.0, max=10000.0)
    return t


def get_pronoun_id(span):
    if len(span) == 1:
        span = list(span)
        if span[0] in PRONOUNS_GROUPS:
            return PRONOUNS_GROUPS[span[0]]
    return -1


def get_category_id(mention, antecedent):
    mention, mention_pronoun_id = mention
    antecedent, antecedent_pronoun_id = antecedent

    if mention_pronoun_id > -1 and antecedent_pronoun_id > -1:
        if mention_pronoun_id == antecedent_pronoun_id:
            return CATEGORIES['pron-pron-comp']
        else:
            return CATEGORIES['pron-pron-no-comp']

    if mention_pronoun_id > -1 or antecedent_pronoun_id > -1:
        return CATEGORIES['pron-ent']

    if mention == antecedent:
        return CATEGORIES['match']

    union = mention.union(antecedent)
    if len(union) == max(len(mention), len(antecedent)):
        return CATEGORIES['contain']

    return CATEGORIES['other']



import json
import logging
import os.path
from collections import defaultdict

import datasets
from datasets import Dataset, Sequence, Value
from tqdm import tqdm

from fastcoref.utilities import util, consts

logger = logging.getLogger(__name__)


def add_speaker_information(tokens, speakers):
    token_to_new_token_map = []
    new_token_to_token_map = []
    new_tokens = []
    last_speaker = None

    for idx, (token, speaker) in enumerate(zip(tokens, speakers)):
        if last_speaker != speaker:
            new_tokens += [consts.SPEAKER_START, speaker, consts.SPEAKER_END]
            new_token_to_token_map += [None, None, None]
            last_speaker = speaker
        token_to_new_token_map.append(len(new_tokens))
        new_token_to_token_map.append(idx)
        new_tokens.append(token)

    return new_tokens, token_to_new_token_map, new_token_to_token_map


def _tokenize(tokenizer, tokens, clusters, speakers):
    new_tokens, token_to_new_token_map, new_token_to_token_map = tokens, list(range(len(tokens))), list(range(len(tokens)))
    if speakers:
        new_tokens, token_to_new_token_map, new_token_to_token_map = add_speaker_information(tokens, speakers)
        for cluster in clusters:
            for start, end in cluster:
                assert tokens[start:end + 1] == new_tokens[token_to_new_token_map[start]:token_to_new_token_map[end] + 1]

    encoded_text = tokenizer(
        new_tokens, add_special_tokens=True, is_split_into_words=True,
        return_length=True, return_attention_mask=False
    )

    # shifting clusters indices to align with bpe tokens
    # align clusters is the reason we can't do it in batches.
    new_clusters = [[(encoded_text.word_to_tokens(token_to_new_token_map[start]).start,
                      encoded_text.word_to_tokens(token_to_new_token_map[end]).end - 1)
                     for start, end in cluster] for cluster in clusters]

    return {'tokens': tokens,
            'input_ids': encoded_text['input_ids'],
            'length': encoded_text['length'][0],

            'gold_clusters': new_clusters,
            # tokens to tokens + speakers
            'new_token_map': new_token_to_token_map,
            # tokens + speakers to bpe
            'subtoken_map': encoded_text.word_ids(),
            }

# TODO: better to do it in batches
def encode(example, tokenizer, nlp):
    if 'tokens' in example and example['tokens']:
        pass
    elif 'text' in example and example['text']:
        clusters = example['clusters']
        spacy_doc = nlp(example['text'])
        example['tokens'] = [tok.text for tok in spacy_doc]

        new_clusters = [[(spacy_doc.char_span(start, end).start,
                          spacy_doc.char_span(start, end).end - 1)
                         for start, end in cluster] for cluster in clusters]
        # verify alignment
        for cluster, new_cluster in zip(clusters, new_clusters):
            for (s1, e1), (s2, e2) in zip(cluster, new_cluster):
                mention = [tok.text for tok in nlp(example['text'][s1:e1])]
                assert mention == example['tokens'][s2:e2 + 1], (mention, example['tokens'][s2:e2 + 1])

        example['clusters'] = new_clusters
    else:
        raise ValueError(f"Example is empty: {example}")

    encoded_example = _tokenize(tokenizer, example['tokens'], example['clusters'], example['speakers'])

    gold_clusters = encoded_example['gold_clusters']
    encoded_example['num_clusters'] = len(gold_clusters) if gold_clusters else 0
    encoded_example['max_cluster_size'] = max(len(c) for c in gold_clusters) if gold_clusters else 0

    return encoded_example


def create(file, tokenizer, nlp):
    def read_jsonlines(file):
        with open(file, 'r') as f:
            for i, line in enumerate(f):
                doc = json.loads(line)
                if "text" not in doc and "tokens" not in doc and "sentences" not in doc:
                    raise ValueError(f'The jsonlines should contains at lt least "text", "sentences" or "tokens" field')

                minimum_doc = {}
                if "doc_key" not in doc:
                    minimum_doc["doc_key"] = str(i)
                else:
                    minimum_doc["doc_key"] = doc["doc_key"]

                if "text" not in doc:
                    minimum_doc["text"] = ""
                else:
                    minimum_doc["text"] = doc["text"]

                if "tokens" in doc:
                    minimum_doc["tokens"] = doc["tokens"]
                elif "sentences" in doc:
                    minimum_doc["tokens"] = util.flatten(doc["sentences"])
                else:
                    minimum_doc["tokens"] = []

                if "speakers" not in doc:
                    minimum_doc["speakers"] = []
                else:
                    minimum_doc["speakers"] = util.flatten(doc["speakers"])

                if "clusters" not in doc:
                    minimum_doc["clusters"] = []
                else:
                    minimum_doc["clusters"] = doc["clusters"]

                yield minimum_doc

    features = datasets.Features(
        {
            "doc_key": Value("string"),
            "text": Value("string"),
            "tokens": Sequence(Value("string")),
            "speakers": Sequence(Value("string")),
            "clusters": Sequence(Sequence(Sequence(Value("int64")))),
        }
    )

    dataset = Dataset.from_generator(read_jsonlines, features=features, gen_kwargs={'file': file})
    dataset = dataset.map(
        encode, batched=False,
        fn_kwargs={'tokenizer': tokenizer, 'nlp': nlp},
        load_from_cache_file=True
    )

    return dataset


# TODO: this function can be implemented much much better, e.g. from_generator
def create_batches(sampler, shuffle=True, cache_dir='cache'):
    # huggingface dataset cannot save tensors. so we will save lists and on train loop transform to tensors.
    batches_dict = defaultdict(lambda: [])

    for i, batch in enumerate(tqdm(sampler, desc="Creating batches for training")):
        for k, v in batch.items():
            batches_dict[k].append(v)

    batches = Dataset.from_dict(batches_dict)

    return batches


import logging
import torch
import math
from fastcoref.utilities.util import pad_clusters


logger = logging.getLogger(__name__)


class LeftOversCollator:
    def __init__(self, tokenizer, device, max_segment_len):
        self.tokenizer = tokenizer
        self.device = device
        self.max_segment_len = max_segment_len

    def __call__(self, batch):
        # pad to the longest doc in the batch
        batch = self.tokenizer.pad(batch)
        batch['leftovers'] = {'input_ids': [], 'attention_mask': []}

        # break down to segment of segment len
        input_ids = [[ids[i:i + self.max_segment_len] for i in range(0, len(ids), self.max_segment_len)]
                     for ids in batch['input_ids']]
        attention_mask = [[mask[i:i + self.max_segment_len] for i in range(0, len(mask), self.max_segment_len)]
                          for mask in batch['attention_mask']]

        # if we have more than 1 segment and the last segment is less than segment_len we have leftovers.
        if len(input_ids[0]) > 1 and len(input_ids[0][-1]) < self.max_segment_len:
            batch['leftovers']['input_ids'] = torch.tensor([ids[-1] for ids in input_ids], device=self.device)
            batch['leftovers']['attention_mask'] = torch.tensor([mask[-1] for mask in attention_mask], device=self.device)

            # remove leftovers from main batch
            input_ids = [ids[:-1] for ids in input_ids]
            attention_mask = [mask[:-1] for mask in attention_mask]

        batch['input_ids'] = torch.tensor(input_ids, device=self.device)
        batch['attention_mask'] = torch.tensor(attention_mask, device=self.device)

        if 'gold_clusters' in batch:
            max_num_clusters, max_max_cluster_size = max(batch['num_clusters']), max(batch['max_cluster_size'])
            if max_num_clusters and max_max_cluster_size:
                padded_clusters = [pad_clusters(cluster, max_num_clusters, max_max_cluster_size) for cluster in
                                   batch['gold_clusters']]
                batch['gold_clusters'] = torch.tensor(padded_clusters, device=self.device)
            else:
                batch['gold_clusters'] = None

        return batch


class PadCollator:
    def __init__(self, tokenizer, device, max_segment_len=512):
        self.tokenizer = tokenizer
        self.device = device
        self.max_segment_len = max_segment_len

    def __call__(self, batch):
        # pad to the longest doc in the batch
        batch = self.tokenizer.pad(batch)

        batch['input_ids'] = torch.tensor(batch['input_ids'], device=self.device)
        batch['attention_mask'] = torch.tensor(batch['attention_mask'], device=self.device)

        if 'gold_clusters' in batch:
            max_num_clusters, max_max_cluster_size = max(batch['num_clusters']), max(batch['max_cluster_size'])
            if max_num_clusters and max_max_cluster_size:
                padded_clusters = [pad_clusters(cluster, max_num_clusters, max_max_cluster_size) for cluster in
                                   batch['gold_clusters']]
                batch['gold_clusters'] = torch.tensor(padded_clusters, device=self.device)
            else:
                batch['gold_clusters'] = None

        return batch


class DynamicBatchSampler:
    def __init__(self, dataset, collator, max_tokens, max_segment_len, max_doc_len=None):
        self.max_tokens = max_tokens
        self.dataset = dataset.sort('length', reverse=False)
        self.collator = collator
        self.max_segment_len = max_segment_len
        self.max_doc_len = max_doc_len

    def __iter__(self):
        batch = []
        per_example_batch_len = 0
        for example in self.dataset:
            if self.max_doc_len is not None and example['length'] > self.max_doc_len:
                logger.info(f'Skipping doc with len {example["length"]}. max_doc_len is {self.max_doc_len}')
                continue
            if not batch:
                per_example_batch_len = self.calc_effective_per_example_batch_len(example['length'])
            elif (len(batch) + 1) * per_example_batch_len > self.max_tokens:
                yield self.collator(batch)
                batch = []
                per_example_batch_len = self.calc_effective_per_example_batch_len(example['length'])
            batch.append(example)
        if len(batch) > 0:
            yield self.collator(batch)

    def calc_effective_per_example_batch_len(self, example_len):
        return math.ceil(example_len / self.max_segment_len) * self.max_segment_len



from setuptools import setup, find_packages

import sys

if sys.version_info < (3,7):
    sys.exit('Python < 3.7 is not supported')

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='fastcoref',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='2.0.3',
    license='MIT',
    author="Shon Otmazgin, Arie Cattan, Yoav Goldberg",
    author_email='shon711@gmail.com',
    packages=['fastcoref', 'fastcoref.coref_models', 'fastcoref.utilities', 'fastcoref.spacy_component'],
    url='https://github.com/shon-otmazgin/fastcoref',
    install_requires=[
        'tqdm>=4.64.0',
        'numpy>=1.21.6',
        'scipy>=1.7.3',
        'spacy>=3.0.6',
        'torch>=1.10.0',
        'transformers>=4.11.3',
        'datasets>=2.5.2',
        'wandb>=0.13.4'
      ],

)



import json
from abc import ABC

import logging
import torch
import transformers
import numpy as np
from tqdm.auto import tqdm
from transformers import AutoConfig, AutoTokenizer
from datasets import Dataset
import spacy
from spacy.cli import download

from fastcoref.coref_models.modeling_fcoref import FCorefModel
from fastcoref.coref_models.modeling_lingmess import LingMessModel
from fastcoref.utilities.util import set_seed, create_mention_to_antecedent, create_clusters, align_to_char_level, encode
from fastcoref.utilities.collate import LeftOversCollator, DynamicBatchSampler, PadCollator

# Setup logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s - %(levelname)s - \t %(message)s',
                    datefmt='%m/%d/%Y %H:%M:%S', level=logging.INFO)


class CorefResult:
    def __init__(self, text, clusters, char_map, reverse_char_map, coref_logit, text_idx):
        self.text = text
        self.clusters = clusters
        self.char_map = char_map
        self.reverse_char_map = reverse_char_map
        self.coref_logit = coref_logit
        self.text_idx = text_idx

    def get_clusters(self, as_strings=True):
        if not as_strings:
            return [[self.char_map[mention][1] for mention in cluster] for cluster in self.clusters]

        return [[self.text[self.char_map[mention][1][0]:self.char_map[mention][1][1]] for mention in cluster]
                for cluster in self.clusters]

    def get_logit(self, span_i, span_j):
        if span_i not in self.reverse_char_map:
            raise ValueError(f'span_i="{self.text[span_i[0]:span_i[1]]}" is not an entity in this model!')
        if span_j not in self.reverse_char_map:
            raise ValueError(f'span_i="{self.text[span_j[0]:span_j[1]]}" is not an entity in this model!')

        span_i_idx = self.reverse_char_map[span_i][0]   # 0 is to get the span index
        span_j_idx = self.reverse_char_map[span_j][0]

        if span_i_idx < span_j_idx:
            return self.coref_logit[span_j_idx, span_i_idx]

        return self.coref_logit[span_i_idx, span_j_idx]

    def __str__(self):
        if len(self.text) > 50:
            text_to_print = f'{self.text[:50]}...'
        else:
            text_to_print = self.text
        return f'CorefResult(text="{text_to_print}", clusters={self.get_clusters()})'

    def __repr__(self):
        return self.__str__()


class CorefModel(ABC):
    def __init__(self, model_name_or_path, coref_class, collator_class, device=None, nlp=None):
        self.model_name_or_path = model_name_or_path
        self.device = device
        self.seed = 42
        self._set_device()

        config = AutoConfig.from_pretrained(self.model_name_or_path)
        self.max_segment_len = config.coref_head['max_segment_len']
        self.max_doc_len = config.coref_head['max_doc_len'] if 'max_doc_len' in config.coref_head else None

        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_name_or_path, use_fast=True,
            add_prefix_space=True, verbose=False
        )

        if collator_class == PadCollator:
            self.collator = PadCollator(tokenizer=self.tokenizer, device=self.device)
        elif collator_class == LeftOversCollator:
            self.collator = LeftOversCollator(
                tokenizer=self.tokenizer, device=self.device,
                max_segment_len=config.coref_head['max_segment_len']
            )
        else:
            raise NotImplementedError(f"Class collator {type(collator_class)} is not supported! "
                                      f"only LeftOversCollator or PadCollator supported")
        if nlp is not None:
            self.nlp = nlp
        else:
            try:
                self.nlp = spacy.load("en_core_web_sm", exclude=["tagger", "parser", "lemmatizer", "ner", "textcat"])
            except OSError:
                # TODO: this is a workaround it is not clear how to add "en_core_web_sm" to setup.py
                download('en_core_web_sm')
                self.nlp = spacy.load("en_core_web_sm", exclude=["tagger", "parser", "lemmatizer", "ner", "textcat"])

        self.model, loading_info = coref_class.from_pretrained(
            self.model_name_or_path, config=config,
            output_loading_info=True
        )
        self.model.to(self.device)

        for key, val in loading_info.items():
            logger.info(f'{key}: {list(set(val) - set(["longformer.embeddings.position_ids"]))}')
        t_params, h_params = [p / 1000000 for p in self.model.num_parameters()]
        logger.info(f'Model Parameters: {t_params + h_params:.1f}M, '
                    f'Transformer: {t_params:.1f}M, Coref head: {h_params:.1f}M')

        set_seed(self)
        transformers.logging.set_verbosity_error()

    def _set_device(self):
        if self.device is None:
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(self.device)
        self.n_gpu = torch.cuda.device_count()

    def _create_dataset(self, texts):
        logger.info(f'Tokenize {len(texts)} texts...')
        # Save original text ordering for later use
        dataset = Dataset.from_dict({'text': texts,'idx':range(len(texts))})
        dataset = dataset.map(
            encode, batched=True, batch_size=10000,
            fn_kwargs={'tokenizer': self.tokenizer, 'nlp': self.nlp}
        )

        return dataset

    def _prepare_batches(self, dataset, max_tokens_in_batch):
        dataloader = DynamicBatchSampler(
            dataset,
            collator=self.collator,
            max_tokens=max_tokens_in_batch,
            max_segment_len=self.max_segment_len,
            max_doc_len=self.max_doc_len
        )

        return dataloader

    def _inference(self, dataloader):
        self.model.eval()
        logger.info(f"***** Running Inference on {len(dataloader.dataset)} texts *****")

        results = []
        with tqdm(desc="Inference", total=len(dataloader.dataset)) as progress_bar:
            for batch in dataloader:
                texts = batch['text']
                subtoken_map = batch['subtoken_map']
                token_to_char = batch['offset_mapping']
                idxs = batch['idx']
                with torch.no_grad():
                    outputs = self.model(batch, return_all_outputs=True)

                outputs_np = tuple(tensor.cpu().numpy() for tensor in outputs)

                span_starts, span_ends, mention_logits, coref_logits = outputs_np
                doc_indices, mention_to_antecedent = create_mention_to_antecedent(span_starts, span_ends, coref_logits)

                for i in range(len(texts)):
                    doc_mention_to_antecedent = mention_to_antecedent[np.nonzero(doc_indices == i)]
                    predicted_clusters = create_clusters(doc_mention_to_antecedent)

                    char_map, reverse_char_map = align_to_char_level(
                        span_starts[i], span_ends[i], token_to_char[i], subtoken_map[i]
                    )

                    res = CorefResult(
                        text=texts[i], clusters=predicted_clusters,
                        char_map=char_map, reverse_char_map=reverse_char_map,
                        coref_logit=coref_logits[i], text_idx=idxs[i]
                    )
                    results.append(res)

                progress_bar.update(n=len(texts))

        return sorted(results, key=lambda res: res.text_idx)

    def predict(self, texts, max_tokens_in_batch=10000, output_file=None):
        is_str = False
        if isinstance(texts, str):
            texts = [texts]
            is_str = True
        if not isinstance(texts, list):
            raise ValueError(f'texts argument expected to be a list of strings, or one single text string. provided {type(texts)}')

        dataset = self._create_dataset(texts=texts)
        dataloader = self._prepare_batches(dataset, max_tokens_in_batch)

        preds = self._inference(dataloader=dataloader)
        if output_file is not None:
            with open(output_file, 'w') as f:
                data = [{'text': p.text,
                         'clusters': p.get_clusters(as_strings=False),
                         'clusters_strings': p.get_clusters(as_strings=True)}
                        for p in preds]
                f.write('\n'.join(map(json.dumps, data)))
        if is_str:
            return preds[0]
        return preds


class FCoref(CorefModel):
    def __init__(self, model_name_or_path='biu-nlp/f-coref', device=None, nlp=None):
        super().__init__(model_name_or_path, FCorefModel, LeftOversCollator, device, nlp)


class LingMessCoref(CorefModel):
    def __init__(self, model_name_or_path='biu-nlp/lingmess-coref', device=None, nlp=None):
        super().__init__(model_name_or_path, LingMessModel, PadCollator, device, nlp)




